@@ -1,9 +1,7 @@
-# coding=utf-8
-#
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2019 David R. MacIver
+# Most of this work is copyright (C) 2013-2021 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -18,32 +16,29 @@
 """This file can approximately be considered the collection of hypothesis going
 to really unreasonable lengths to produce pretty output."""
 
-
-from __future__ import absolute_import, division, print_function
-
 import ast
 import hashlib
 import inspect
+import os
 import re
-import tokenize
 import types
-import uuid
 from functools import wraps
+from tokenize import detect_encoding
 from types import ModuleType
+from typing import Callable, TypeVar
 
 from hypothesis.internal.compat import (
-    ARG_NAME_ATTRIBUTE,
-    getfullargspec,
-    hrange,
-    isidentifier,
+    is_typed_named_tuple,
     qualname,
     str_to_bytes,
-    to_str,
     to_unicode,
     update_code_location,
 )
 from hypothesis.vendor.pretty import pretty
 
+C = TypeVar("C", bound=Callable)
+READTHEDOCS = os.environ.get("READTHEDOCS", None) == "True"
+
 
 def fully_qualified_name(f):
     """Returns a unique identifier for f pointing to the module it was defined
@@ -55,17 +50,15 @@ def fully_qualified_name(f):
 
 
 def is_mock(obj):
-    """Determine if the given argument is a mock type.
+    """Determine if the given argument is a mock type."""
 
-    We want to be able to detect these when dealing with various test
-    args. As they are sneaky and can look like almost anything else,
-    we'll check this by looking for random attributes.  This is more
-    robust than looking for types.
-    """
-    for _ in range(10):
-        if not hasattr(obj, str(uuid.uuid4())):
-            return False
-    return True
+    # We want to be able to detect these when dealing with various test
+    # args. As they are sneaky and can look like almost anything else,
+    # we'll check this by looking for an attribute with a name that it's really
+    # unlikely to implement accidentally, and that anyone who implements it
+    # deliberately should know what they're doing. This is more robust than
+    # looking for types.
+    return hasattr(obj, "hypothesis_internal_is_this_a_mock_check")
 
 
 def function_digest(function):
@@ -75,11 +68,10 @@ def function_digest(function):
 
     No guarantee of uniqueness though it usually will be.
     """
-    hasher = hashlib.md5()
+    hasher = hashlib.sha384()
     try:
         hasher.update(to_unicode(inspect.getsource(function)).encode("utf-8"))
-    # Different errors on different versions of python. What fun.
-    except (OSError, IOError, TypeError):
+    except (OSError, TypeError):
         pass
     try:
         hasher.update(str_to_bytes(function.__name__))
@@ -90,7 +82,7 @@ def function_digest(function):
     except AttributeError:
         pass
     try:
-        hasher.update(str_to_bytes(repr(getfullargspec(function))))
+        hasher.update(str_to_bytes(repr(inspect.getfullargspec(function))))
     except TypeError:
         pass
     try:
@@ -100,21 +92,6 @@ def function_digest(function):
     return hasher.digest()
 
 
-def is_typed_named_tuple(cls):
-    """Return True if cls is probably a subtype of `typing.NamedTuple`.
-
-    Unfortunately types created with `class T(NamedTuple):` actually
-    subclass `tuple` directly rather than NamedTuple.  This is annoying,
-    and means we just have to hope that nobody defines a different tuple
-    subclass with similar attributes.
-    """
-    return (
-        issubclass(cls, tuple)
-        and hasattr(cls, "_fields")
-        and hasattr(cls, "_field_types")
-    )
-
-
 def required_args(target, args=(), kwargs=()):
     """Return a set of names of required args to target that were not supplied
     in args or kwargs.
@@ -128,13 +105,18 @@ def required_args(target, args=(), kwargs=()):
     if inspect.isclass(target) and is_typed_named_tuple(target):
         provided = set(kwargs) | set(target._fields[: len(args)])
         return set(target._fields) - provided
-    # Then we try to do the right thing with getfullargspec
+    # Then we try to do the right thing with inspect.getfullargspec
+    # Note that for classes we inspect the __init__ method, *unless* the class
+    # has an explicit __signature__ attribute.  This allows us to support
+    # runtime-generated constraints on **kwargs, as for e.g. Pydantic models.
     try:
-        spec = getfullargspec(
-            getattr(target, "__init__", target) if inspect.isclass(target) else target
+        spec = inspect.getfullargspec(
+            getattr(target, "__init__", target)
+            if inspect.isclass(target) and not hasattr(target, "__signature__")
+            else target
         )
     except TypeError:  # pragma: no cover
-        return None
+        return set()
     # self appears in the argspec of __init__ and bound methods, but it's an
     # error to explicitly supply it - so we might skip the first argument.
     skip_self = int(inspect.isclass(target) or inspect.ismethod(target))
@@ -151,11 +133,10 @@ def required_args(target, args=(), kwargs=()):
 
 def convert_keyword_arguments(function, args, kwargs):
     """Returns a pair of a tuple and a dictionary which would be equivalent
-    passed as positional and keyword args to the function. Unless function has.
-
-    **kwargs the dictionary will always be empty.
+    passed as positional and keyword args to the function. Unless function has
+    kwonlyargs or **kwargs the dictionary will always be empty.
     """
-    argspec = getfullargspec(function)
+    argspec = inspect.getfullargspec(function)
     new_args = []
     kwargs = dict(kwargs)
 
@@ -169,7 +150,7 @@ def convert_keyword_arguments(function, args, kwargs):
 
     n = max(len(args), len(argspec.args))
 
-    for i in hrange(n):
+    for i in range(n):
         if i < len(args):
             new_args.append(args[i])
         else:
@@ -179,7 +160,7 @@ def convert_keyword_arguments(function, args, kwargs):
             elif arg_name in defaults:
                 new_args.append(defaults[arg_name])
             else:
-                raise TypeError("No value provided for argument %r" % (arg_name))
+                raise TypeError(f"No value provided for argument {arg_name!r}")
 
     if kwargs and not (argspec.varkw or argspec.kwonlyargs):
         if len(kwargs) > 1:
@@ -202,7 +183,7 @@ def convert_positional_arguments(function, args, kwargs):
 
     new_args will only be non-empty if function has a variadic argument.
     """
-    argspec = getfullargspec(function)
+    argspec = inspect.getfullargspec(function)
     new_kwargs = dict(argspec.kwonlydefaults or {})
     new_kwargs.update(kwargs)
     if not argspec.varkw:
@@ -213,29 +194,28 @@ def convert_positional_arguments(function, args, kwargs):
                     % (function.__name__, k)
                 )
     if len(args) < len(argspec.args):
-        for i in hrange(len(args), len(argspec.args) - len(argspec.defaults or ())):
+        for i in range(len(args), len(argspec.args) - len(argspec.defaults or ())):
             if argspec.args[i] not in kwargs:
-                raise TypeError(
-                    "No value provided for argument %s" % (argspec.args[i],)
-                )
+                raise TypeError(f"No value provided for argument {argspec.args[i]}")
     for kw in argspec.kwonlyargs:
         if kw not in new_kwargs:
-            raise TypeError("No value provided for argument %s" % kw)
+            raise TypeError(f"No value provided for argument {kw}")
 
     if len(args) > len(argspec.args) and not argspec.varargs:
         raise TypeError(
-            "%s() takes at most %d positional arguments (%d given)"
-            % (function.__name__, len(argspec.args), len(args))
+            "{}() takes at most {} positional arguments ({} given)".format(
+                function.__name__, len(argspec.args), len(args)
+            )
         )
 
     for arg, name in zip(args, argspec.args):
         if name in new_kwargs:
             raise TypeError(
-                "%s() got multiple values for keyword argument %r"
-                % (function.__name__, name)
+                f"{function.__name__}() got multiple values for keyword argument {name!r}"
             )
         else:
             new_kwargs[name] = arg
+
     return (tuple(args[len(argspec.args) :]), new_kwargs)
 
 
@@ -252,7 +232,7 @@ def extract_all_lambdas(tree):
 
 
 def args_for_lambda_ast(l):
-    return [getattr(n, ARG_NAME_ATTRIBUTE) for n in l.args.args]
+    return [n.arg for n in l.args.args]
 
 
 LINE_CONTINUATION = re.compile(r"\\\n")
@@ -269,19 +249,11 @@ def extract_lambda_source(f):
     This is not a good function and I am sorry for it. Forgive me my
     sins, oh lord
     """
-    argspec = getfullargspec(f)
+    argspec = inspect.getfullargspec(f)
     arg_strings = []
-    # In Python 2 you can have destructuring arguments to functions. This
-    # results in an argspec with non-string values. I'm not very interested in
-    # handling these properly, but it's important to not crash on them.
-    bad_lambda = False
     for a in argspec.args:
-        if isinstance(a, (tuple, list)):  # pragma: no cover
-            arg_strings.append("(%s)" % (", ".join(a),))
-            bad_lambda = True
-        else:
-            assert isinstance(a, str)
-            arg_strings.append(a)
+        assert isinstance(a, str)
+        arg_strings.append(a)
     if argspec.varargs:
         arg_strings.append("*" + argspec.varargs)
     elif argspec.kwonlyargs:
@@ -289,19 +261,17 @@ def extract_lambda_source(f):
     for a in argspec.kwonlyargs or []:
         default = (argspec.kwonlydefaults or {}).get(a)
         if default:
-            arg_strings.append("{}={}".format(a, default))
+            arg_strings.append(f"{a}={default}")
         else:
             arg_strings.append(a)
 
     if arg_strings:
-        if_confused = "lambda %s: <unknown>" % (", ".join(arg_strings),)
+        if_confused = "lambda {}: <unknown>".format(", ".join(arg_strings))
     else:
         if_confused = "lambda: <unknown>"
-    if bad_lambda:  # pragma: no cover
-        return if_confused
     try:
         source = inspect.getsource(f)
-    except IOError:
+    except OSError:
         return if_confused
 
     source = LINE_CONTINUATION.sub(" ", source)
@@ -314,7 +284,7 @@ def extract_lambda_source(f):
     try:
         tree = ast.parse(source)
     except SyntaxError:
-        for i in hrange(len(source) - 1, len("lambda"), -1):
+        for i in range(len(source) - 1, len("lambda"), -1):
             prefix = source[:i]
             if "lambda" not in prefix:
                 break
@@ -324,22 +294,22 @@ def extract_lambda_source(f):
                 break
             except SyntaxError:
                 continue
-    if tree is None:
-        if source.startswith("@"):
-            # This will always eventually find a valid expression because
-            # the decorator must be a valid Python function call, so will
-            # eventually be syntactically valid and break out of the loop. Thus
-            # this loop can never terminate normally, so a no branch pragma is
-            # appropriate.
-            for i in hrange(len(source) + 1):  # pragma: no branch
-                p = source[1:i]
-                if "lambda" in p:
-                    try:
-                        tree = ast.parse(p)
-                        source = p
-                        break
-                    except SyntaxError:
-                        pass
+    if tree is None and source.startswith("@"):
+        # This will always eventually find a valid expression because
+        # the decorator must be a valid Python function call, so will
+        # eventually be syntactically valid and break out of the loop.
+        # Thus, this loop can never terminate normally.
+        for i in range(len(source) + 1):
+            p = source[1:i]
+            if "lambda" in p:
+                try:
+                    tree = ast.parse(p)
+                    source = p
+                    break
+                except SyntaxError:
+                    pass
+        else:
+            raise NotImplementedError("expected to be unreachable")
 
     if tree is None:
         return if_confused
@@ -368,21 +338,14 @@ def extract_lambda_source(f):
     # an OSError.  Or if `f` is a built-in function, in which case we get a
     # TypeError.  In both cases, fall back to splitting the Unicode string.
     # It's not perfect, but it's the best we can do.
-    #
-    # Note 2: You can only detect the encoding with `tokenize.detect_encoding`
-    # in Python 3.2 or later.  But that's okay, because the only version that
-    # affects for us is Python 2.7, and 2.7 doesn't support non-ASCII identifiers:
-    # https://www.python.org/dev/peps/pep-3131/. In this case we'll get an
-    # AttributeError.
-    #
     try:
         with open(inspect.getsourcefile(f), "rb") as src_f:
-            encoding, _ = tokenize.detect_encoding(src_f.readline)
+            encoding, _ = detect_encoding(src_f.readline)
 
         source_bytes = source.encode(encoding)
         source_bytes = source_bytes[lambda_ast.col_offset :].strip()
         source = source_bytes.decode(encoding)
-    except (AttributeError, OSError, TypeError):
+    except (OSError, TypeError):
         source = source[lambda_ast.col_offset :].strip()
 
     # This ValueError can be thrown in Python 3 if:
@@ -400,7 +363,7 @@ def extract_lambda_source(f):
     except ValueError:
         return if_confused
 
-    for i in hrange(len(source), len("lambda"), -1):  # pragma: no branch
+    for i in range(len(source), len("lambda"), -1):  # pragma: no branch
         try:
             parsed = ast.parse(source[:i])
             assert len(parsed.body) == 1
@@ -427,10 +390,12 @@ def get_pretty_function_description(f):
     name = f.__name__
     if name == "<lambda>":
         return extract_lambda_source(f)
-    elif isinstance(f, types.MethodType):
+    elif isinstance(f, (types.MethodType, types.BuiltinMethodType)):
         self = f.__self__
-        if not (self is None or inspect.isclass(self)):
-            return "%r.%s" % (self, name)
+        # Some objects, like `builtins.abs` are of BuiltinMethodType but have
+        # their module as __self__.  This might include c-extensions generally?
+        if not (self is None or inspect.isclass(self) or inspect.ismodule(self)):
+            return f"{self!r}.{name}"
     return name
 
 
@@ -440,23 +405,23 @@ def nicerepr(v):
     elif isinstance(v, type):
         return v.__name__
     else:
-        return to_str(pretty(v))
+        return pretty(v)
 
 
 def arg_string(f, args, kwargs, reorder=True):
     if reorder:
         args, kwargs = convert_positional_arguments(f, args, kwargs)
 
-    argspec = getfullargspec(f)
+    argspec = inspect.getfullargspec(f)
 
     bits = []
 
     for a in argspec.args:
         if a in kwargs:
-            bits.append("%s=%s" % (a, nicerepr(kwargs.pop(a))))
+            bits.append("{}={}".format(a, nicerepr(kwargs.pop(a))))
     if kwargs:
         for a in sorted(kwargs):
-            bits.append("%s=%s" % (a, nicerepr(kwargs[a])))
+            bits.append("{}={}".format(a, nicerepr(kwargs[a])))
 
     return ", ".join([nicerepr(x) for x in args] + bits)
 
@@ -468,11 +433,11 @@ def unbind_method(f):
 
 
 def check_valid_identifier(identifier):
-    if not isidentifier(identifier):
-        raise ValueError("%r is not a valid python identifier" % (identifier,))
+    if not identifier.isidentifier():
+        raise ValueError(f"{identifier!r} is not a valid python identifier")
 
 
-eval_cache = {}  # type: dict
+eval_cache: dict = {}
 
 
 def source_exec_as_module(source):
@@ -483,7 +448,7 @@ def source_exec_as_module(source):
 
     result = ModuleType(
         "hypothesis_temporary_module_%s"
-        % (hashlib.sha1(str_to_bytes(source)).hexdigest(),)
+        % (hashlib.sha384(str_to_bytes(source)).hexdigest(),)
     )
     assert isinstance(source, str)
     exec(source, result.__dict__)
@@ -517,7 +482,7 @@ def define_function_signature(name, docstring, argspec):
         for a in argspec.args[:-n_defaults]:
             parts.append(a)
         for a in argspec.args[-n_defaults:]:
-            parts.append("%s=not_set" % (a,))
+            parts.append(f"{a}=not_set")
     else:
         parts = list(argspec.args)
     used_names = list(argspec.args) + list(argspec.kwonlyargs)
@@ -527,7 +492,7 @@ def define_function_signature(name, docstring, argspec):
         check_valid_identifier(a)
 
     def accept(f):
-        fargspec = getfullargspec(f)
+        fargspec = inspect.getfullargspec(f)
         must_pass_as_kwargs = []
         invocation_parts = []
         for a in argspec.args:
@@ -542,12 +507,12 @@ def define_function_signature(name, docstring, argspec):
         elif argspec.kwonlyargs:
             parts.append("*")
         for k in must_pass_as_kwargs:
-            invocation_parts.append("%(k)s=%(k)s" % {"k": k})
+            invocation_parts.append(f"{k}={k}")
 
         for k in argspec.kwonlyargs:
-            invocation_parts.append("%(k)s=%(k)s" % {"k": k})
+            invocation_parts.append(f"{k}={k}")
             if k in (argspec.kwonlydefaults or []):
-                parts.append("%(k)s=not_set" % {"k": k})
+                parts.append(f"{k}=not_set")
             else:
                 parts.append(k)
         if argspec.varkw:
@@ -555,9 +520,7 @@ def define_function_signature(name, docstring, argspec):
             parts.append("**" + argspec.varkw)
             invocation_parts.append("**" + argspec.varkw)
 
-        candidate_names = ["f"] + [
-            "f_%d" % (i,) for i in hrange(1, len(used_names) + 2)
-        ]
+        candidate_names = ["f"] + [f"f_{i}" for i in range(1, len(used_names) + 2)]
 
         for funcname in candidate_names:  # pragma: no branch
             if funcname not in used_names:
@@ -613,7 +576,7 @@ def proxies(target):
                 define_function_signature(
                     target.__name__.replace("<lambda>", "_lambda_"),
                     target.__doc__,
-                    getfullargspec(target),
+                    inspect.getfullargspec(target),
                 )(proxy)
             )
         )
