@@ -1,7 +1,7 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2020 David R. MacIver
+# Most of this work is copyright (C) 2013-2021 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -36,14 +36,9 @@ from hypothesis._settings import HealthCheck, Verbosity, settings as Settings
 from hypothesis.control import current_build_context
 from hypothesis.core import given
 from hypothesis.errors import InvalidArgument, InvalidDefinition
+from hypothesis.internal.compat import qualname
 from hypothesis.internal.conjecture import utils as cu
-from hypothesis.internal.reflection import (
-    deprecated_posargs,
-    function_digest,
-    nicerepr,
-    proxies,
-    qualname,
-)
+from hypothesis.internal.reflection import function_digest, nicerepr, proxies
 from hypothesis.internal.validation import check_type
 from hypothesis.reporting import current_verbosity, report
 from hypothesis.strategies._internal.featureflags import FeatureStrategy
@@ -71,7 +66,6 @@ class TestCaseProperty:  # pragma: no cover
         raise AttributeError("Cannot delete TestCase")
 
 
-@deprecated_posargs
 def run_state_machine_as_test(state_machine_factory, *, settings=None):
     """Run a state machine definition as a test, either silently doing nothing
     or printing a minimal breaking program and raising an exception.
@@ -101,7 +95,7 @@ def run_state_machine_as_test(state_machine_factory, *, settings=None):
         )
         try:
             if print_steps:
-                report("state = %s()" % (machine.__class__.__name__,))
+                report(f"state = {machine.__class__.__name__}()")
             machine.check_invariants()
             max_steps = settings.stateful_step_count
             steps_run = 0
@@ -185,8 +179,8 @@ def run_state_machine_as_test(state_machine_factory, *, settings=None):
             machine.teardown()
 
     # Use a machine digest to identify stateful tests in the example database
-    run_state_machine.hypothesis.inner_test._hypothesis_internal_add_digest = function_digest(
-        state_machine_factory
+    run_state_machine.hypothesis.inner_test._hypothesis_internal_add_digest = (
+        function_digest(state_machine_factory)
     )
     # Copy some attributes so @seed and @reproduce_failure "just work"
     run_state_machine._hypothesis_internal_use_seed = getattr(
@@ -223,18 +217,16 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
     executed.
     """
 
-    _rules_per_class = {}  # type: Dict[type, List[classmethod]]
-    _invariants_per_class = {}  # type: Dict[type, List[classmethod]]
-    _base_rules_per_class = {}  # type: Dict[type, List[classmethod]]
-    _initializers_per_class = {}  # type: Dict[type, List[classmethod]]
-    _base_initializers_per_class = {}  # type: Dict[type, List[classmethod]]
+    _rules_per_class: Dict[type, List[classmethod]] = {}
+    _invariants_per_class: Dict[type, List[classmethod]] = {}
+    _initializers_per_class: Dict[type, List[classmethod]] = {}
 
     def __init__(self):
         if not self.rules():
-            raise InvalidDefinition("Type %s defines no rules" % (type(self).__name__,))
-        self.bundles = {}  # type: Dict[str, list]
+            raise InvalidDefinition(f"Type {type(self).__name__} defines no rules")
+        self.bundles: Dict[str, list] = {}
         self.name_counter = 1
-        self.names_to_values = {}  # type: Dict[str, Any]
+        self.names_to_values: Dict[str, Any] = {}
         self.__stream = StringIO()
         self.__printer = RepresentationPrinter(self.__stream)
         self._initialize_rules_to_run = copy(self.initialize_rules())
@@ -253,17 +245,17 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
         return self.__stream.getvalue()
 
     def __repr__(self):
-        return "%s(%s)" % (type(self).__name__, nicerepr(self.bundles))
+        return "{}({})".format(type(self).__name__, nicerepr(self.bundles))
 
     def _new_name(self):
-        result = "v%d" % (self.name_counter,)
+        result = f"v{self.name_counter}"
         self.name_counter += 1
         return result
 
     def _last_names(self, n):
         assert self.name_counter > n
         count = self.name_counter
-        return ["v%d" % (i,) for i in range(count - n, count)]
+        return [f"v{i}" for i in range(count - n, count)]
 
     def bundle(self, name):
         return self.bundles.setdefault(name, [])
@@ -275,13 +267,11 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
         except KeyError:
             pass
 
+        cls._initializers_per_class[cls] = []
         for _, v in inspect.getmembers(cls):
             r = getattr(v, INITIALIZE_RULE_MARKER, None)
             if r is not None:
-                cls.define_initialize_rule(
-                    r.targets, r.function, r.arguments, r.precondition
-                )
-        cls._initializers_per_class[cls] = cls._base_initializers_per_class.pop(cls, [])
+                cls._initializers_per_class[cls].append(r)
         return cls._initializers_per_class[cls]
 
     @classmethod
@@ -291,11 +281,11 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
         except KeyError:
             pass
 
+        cls._rules_per_class[cls] = []
         for _, v in inspect.getmembers(cls):
             r = getattr(v, RULE_MARKER, None)
             if r is not None:
-                cls.define_rule(r.targets, r.function, r.arguments, r.precondition)
-        cls._rules_per_class[cls] = cls._base_rules_per_class.pop(cls, [])
+                cls._rules_per_class[cls].append(r)
         return cls._rules_per_class[cls]
 
     @classmethod
@@ -313,30 +303,6 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
         cls._invariants_per_class[cls] = target
         return cls._invariants_per_class[cls]
 
-    @classmethod
-    def define_initialize_rule(cls, targets, function, arguments, precondition=None):
-        converted_arguments = {}
-        for k, v in arguments.items():
-            converted_arguments[k] = v
-        if cls in cls._initializers_per_class:
-            target = cls._initializers_per_class[cls]
-        else:
-            target = cls._base_initializers_per_class.setdefault(cls, [])
-
-        return target.append(Rule(targets, function, converted_arguments, precondition))
-
-    @classmethod
-    def define_rule(cls, targets, function, arguments, precondition=None):
-        converted_arguments = {}
-        for k, v in arguments.items():
-            converted_arguments[k] = v
-        if cls in cls._rules_per_class:
-            target = cls._rules_per_class[cls]
-        else:
-            target = cls._base_rules_per_class.setdefault(cls, [])
-
-        return target.append(Rule(targets, function, converted_arguments, precondition))
-
     def _print_step(self, rule, data, result):
         self.step_count = getattr(self, "step_count", 0) + 1
         # If the step has target bundles, and the result is a MultipleResults
@@ -345,14 +311,12 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
             n_output_vars = len(result.values)
         else:
             n_output_vars = 1
-        output_assignment = (
-            "%s = " % (", ".join(self._last_names(n_output_vars)),)
-            if rule.targets and n_output_vars >= 1
-            else ""
-        )
+        if rule.targets and n_output_vars >= 1:
+            output_assignment = ", ".join(self._last_names(n_output_vars)) + " = "
+        else:
+            output_assignment = ""
         report(
-            "%sstate.%s(%s)"
-            % (
+            "{}state.{}({})".format(
                 output_assignment,
                 rule.function.__name__,
                 ", ".join("%s=%s" % kv for kv in data.items()),
@@ -370,7 +334,9 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
 
     def check_invariants(self):
         for invar in self.invariants():
-            if invar.precondition and not invar.precondition(self):
+            if self._initialize_rules_to_run and not invar.check_during_init:
+                continue
+            if not all(precond(self) for precond in invar.preconditions):
                 continue
             invar.function(self)
 
@@ -404,7 +370,7 @@ class Rule:
     targets = attr.ib()
     function = attr.ib(repr=qualname)
     arguments = attr.ib()
-    precondition = attr.ib()
+    preconditions = attr.ib()
     bundles = attr.ib(init=False)
 
     def __attrs_post_init__(self):
@@ -458,8 +424,8 @@ class Bundle(SearchStrategy):
     def __repr__(self):
         consume = self.__reference_strategy.consume
         if consume is False:
-            return "Bundle(name=%r)" % (self.name,)
-        return "Bundle(name=%r, consume=%r)" % (self.name, consume)
+            return f"Bundle(name={self.name!r})"
+        return f"Bundle(name={self.name!r}, consume={consume!r})"
 
     def calc_is_empty(self, recur):
         # We assume that a bundle will grow over time
@@ -545,11 +511,10 @@ def _convert_targets(targets, target):
 
 RULE_MARKER = "hypothesis_stateful_rule"
 INITIALIZE_RULE_MARKER = "hypothesis_stateful_initialize_rule"
-PRECONDITION_MARKER = "hypothesis_stateful_precondition"
+PRECONDITIONS_MARKER = "hypothesis_stateful_preconditions"
 INVARIANT_MARKER = "hypothesis_stateful_invariant"
 
 
-@deprecated_posargs
 def rule(*, targets=(), target=None, **kwargs):
     """Decorator for RuleBasedStateMachine. Any name present in target or
     targets will define where the end result of this function should go. If
@@ -575,18 +540,23 @@ def rule(*, targets=(), target=None, **kwargs):
         check_strategy(v, name=k)
 
     def accept(f):
+        if getattr(f, INVARIANT_MARKER, None):
+            raise InvalidDefinition(
+                "A function cannot be used for both a rule and an invariant.",
+                Settings.default,
+            )
         existing_rule = getattr(f, RULE_MARKER, None)
         existing_initialize_rule = getattr(f, INITIALIZE_RULE_MARKER, None)
         if existing_rule is not None or existing_initialize_rule is not None:
             raise InvalidDefinition(
                 "A function cannot be used for two distinct rules. ", Settings.default
             )
-        precondition = getattr(f, PRECONDITION_MARKER, None)
+        preconditions = getattr(f, PRECONDITIONS_MARKER, ())
         rule = Rule(
             targets=converted_targets,
             arguments=kwargs,
             function=f,
-            precondition=precondition,
+            preconditions=preconditions,
         )
 
         @proxies(f)
@@ -599,28 +569,33 @@ def rule(*, targets=(), target=None, **kwargs):
     return accept
 
 
-@deprecated_posargs
 def initialize(*, targets=(), target=None, **kwargs):
     """Decorator for RuleBasedStateMachine.
 
-    An initialize decorator behaves like a rule, but the decorated
-    method is called at most once in a run. All initialize decorated
-    methods will be called before any rule decorated methods, in an
-    arbitrary order.
+    An initialize decorator behaves like a rule, but all ``@initialize()`` decorated
+    methods will be called before any ``@rule()`` decorated methods, in an arbitrary
+    order.  Each ``@initialize()`` method will be called exactly once per run, unless
+    one raises an exception - after which only the ``.teardown()`` method will be run.
+    ``@initialize()`` methods may not have preconditions.
     """
     converted_targets = _convert_targets(targets, target)
     for k, v in kwargs.items():
         check_strategy(v, name=k)
 
     def accept(f):
+        if getattr(f, INVARIANT_MARKER, None):
+            raise InvalidDefinition(
+                "A function cannot be used for both a rule and an invariant.",
+                Settings.default,
+            )
         existing_rule = getattr(f, RULE_MARKER, None)
         existing_initialize_rule = getattr(f, INITIALIZE_RULE_MARKER, None)
         if existing_rule is not None or existing_initialize_rule is not None:
             raise InvalidDefinition(
                 "A function cannot be used for two distinct rules. ", Settings.default
             )
-        precondition = getattr(f, PRECONDITION_MARKER, None)
-        if precondition:
+        preconditions = getattr(f, PRECONDITIONS_MARKER, ())
+        if preconditions:
             raise InvalidDefinition(
                 "An initialization rule cannot have a precondition. ", Settings.default
             )
@@ -628,7 +603,7 @@ def initialize(*, targets=(), target=None, **kwargs):
             targets=converted_targets,
             arguments=kwargs,
             function=f,
-            precondition=precondition,
+            preconditions=preconditions,
         )
 
         @proxies(f)
@@ -649,7 +624,8 @@ class VarReference:
 def precondition(precond):
     """Decorator to apply a precondition for rules in a RuleBasedStateMachine.
     Specifies a precondition for a rule to be considered as a valid step in the
-    state machine. The given function will be called with the instance of
+    state machine, which is more efficient than using :func:`~hypothesis.assume`
+    within the rule.  The ``precond`` function will be called with the instance of
     RuleBasedStateMachine and should return True or False. Usually it will need
     to look at attributes on that instance.
 
@@ -663,8 +639,9 @@ def precondition(precond):
             def divide_with(self, numerator):
                 self.state = numerator / self.state
 
-    This is better than using assume in your rule since more valid rules
-    should be able to be run.
+    If multiple preconditions are applied to a single rule, it is only considered
+    a valid step when all of them return True.  Preconditions may be applied to
+    invariants as well as rules.
     """
 
     def decorator(f):
@@ -679,21 +656,23 @@ def precondition(precond):
             )
 
         rule = getattr(f, RULE_MARKER, None)
-        if rule is None:
-            setattr(precondition_wrapper, PRECONDITION_MARKER, precond)
-        else:
-            new_rule = Rule(
-                targets=rule.targets,
-                arguments=rule.arguments,
-                function=rule.function,
-                precondition=precond,
-            )
-            setattr(precondition_wrapper, RULE_MARKER, new_rule)
-
         invariant = getattr(f, INVARIANT_MARKER, None)
-        if invariant is not None:
-            new_invariant = Invariant(function=invariant.function, precondition=precond)
+        if rule is not None:
+            assert invariant is None
+            new_rule = attr.evolve(rule, preconditions=rule.preconditions + (precond,))
+            setattr(precondition_wrapper, RULE_MARKER, new_rule)
+        elif invariant is not None:
+            assert rule is None
+            new_invariant = attr.evolve(
+                invariant, preconditions=invariant.preconditions + (precond,)
+            )
             setattr(precondition_wrapper, INVARIANT_MARKER, new_invariant)
+        else:
+            setattr(
+                precondition_wrapper,
+                PRECONDITIONS_MARKER,
+                getattr(f, PRECONDITIONS_MARKER, ()) + (precond,),
+            )
 
         return precondition_wrapper
 
@@ -703,10 +682,11 @@ def precondition(precond):
 @attr.s()
 class Invariant:
     function = attr.ib()
-    precondition = attr.ib()
+    preconditions = attr.ib()
+    check_during_init = attr.ib()
 
 
-def invariant():
+def invariant(*, check_during_init=False):
     """Decorator to apply an invariant for rules in a RuleBasedStateMachine.
     The decorated function will be run after every rule and can raise an
     exception to indicate failed invariants.
@@ -719,23 +699,38 @@ def invariant():
             @invariant()
             def is_nonzero(self):
                 assert self.state != 0
+
+    By default, invariants are only checked after all
+    :func:`@initialize() <hypothesis.stateful.initialize>` rules have been run.
+    Pass ``check_during_init=True`` for invariants which can also be checked
+    during initialization.
     """
+    check_type(bool, check_during_init, "check_during_init")
 
     def accept(f):
+        if getattr(f, RULE_MARKER, None) or getattr(f, INITIALIZE_RULE_MARKER, None):
+            raise InvalidDefinition(
+                "A function cannot be used for both a rule and an invariant.",
+                Settings.default,
+            )
         existing_invariant = getattr(f, INVARIANT_MARKER, None)
         if existing_invariant is not None:
             raise InvalidDefinition(
                 "A function cannot be used for two distinct invariants.",
                 Settings.default,
             )
-        precondition = getattr(f, PRECONDITION_MARKER, None)
-        rule = Invariant(function=f, precondition=precondition)
+        preconditions = getattr(f, PRECONDITIONS_MARKER, ())
+        invar = Invariant(
+            function=f,
+            preconditions=preconditions,
+            check_during_init=check_during_init,
+        )
 
         @proxies(f)
         def invariant_wrapper(*args, **kwargs):
             return f(*args, **kwargs)
 
-        setattr(invariant_wrapper, INVARIANT_MARKER, rule)
+        setattr(invariant_wrapper, INVARIANT_MARKER, invar)
         return invariant_wrapper
 
     return accept
@@ -751,7 +746,7 @@ class RuleStrategy(SearchStrategy):
         self.rules = list(machine.rules())
 
         self.enabled_rules_strategy = st.shared(
-            FeatureStrategy(), key=("enabled rules", machine),
+            FeatureStrategy(), key=("enabled rules", machine)
         )
 
         # The order is a bit arbitrary. Primarily we're trying to group rules
@@ -770,14 +765,14 @@ class RuleStrategy(SearchStrategy):
         )
 
     def __repr__(self):
-        return "%s(machine=%s({...}))" % (
+        return "{}(machine={}({{...}}))".format(
             self.__class__.__name__,
             self.machine.__class__.__name__,
         )
 
     def do_draw(self, data):
         if not any(self.is_valid(rule) for rule in self.rules):
-            msg = "No progress can be made from state %r" % (self.machine,)
+            msg = f"No progress can be made from state {self.machine!r}"
             raise InvalidDefinition(msg) from None
 
         feature_flags = data.draw(self.enabled_rules_strategy)
@@ -797,7 +792,7 @@ class RuleStrategy(SearchStrategy):
         return (rule, data.draw(rule.arguments_strategy))
 
     def is_valid(self, rule):
-        if rule.precondition and not rule.precondition(self.machine):
+        if not all(precond(self.machine) for precond in rule.preconditions):
             return False
 
         for b in rule.bundles:
